module main

import net.http
import log
import os
import math

interface Provider {
	install(version string, install_to string) !
}

fn Provider.for_browser(browser_type BrowserType, cache Cache) !Provider {
	return match browser_type {
		.uchromium { Provider(UChromium.new(cache)!) }
	}
}

const uchromium_repo = 'https://github.com/ungoogled-software/ungoogled-chromium-binaries.git'

struct UChromiumVersion {
	url    string
	md5    string
	sha1   string
	sha256 string
}

fn UChromiumVersion.from_file(file string) !UChromiumVersion {
	lines := os.read_lines(file) or { return error('failed to read file ${file}: ${err}') }
	mut data := map[string]string{}

	for line in lines {
		parts := line.split_any(' =').filter(it != '')
		if parts.len != 2 {
			continue
		}

		data[parts[0]] = parts[1] or { '' }
	}

	log.debug('data: ${data}')

	return UChromiumVersion{
		url:    data['url']
		md5:    data['md5']
		sha1:   data['sha1']
		sha256: data['sha256']
	}
}

@[noinit]
struct UChromium {
	cache Cache
}

fn UChromium.new(cache Cache) !UChromium {
	return UChromium{
		cache: cache
	}
}

fn (u UChromium) repo_path() !string {
	return u.cache.dir_no_create('uchromium-repo')
}

fn (u UChromium) set_repo() ! {
	repo_path := u.repo_path()!
	cmd := match os.is_dir(repo_path) {
		false {
			log.info('cloning ${uchromium_repo} to ${repo_path}')
			'git clone ${uchromium_repo} ${repo_path}'
		}
		true {
			log.info('updating ${repo_path}')
			'git -C ${repo_path} pull'
		}
	}

	res := os.execute_opt(cmd) or {
		return error('failed to sync ${uchromium_repo} to ${repo_path}: ${err}')
	}

	log.debug('res: ${res}')

	return
}

fn (u UChromium) find_version(version string) !UChromiumVersion {
	base_dir := os.join_path(u.repo_path()!, 'config', 'platforms', 'appimage', '64bit')
	log.debug('base_dir: ${base_dir}')

	mut versions := os.glob(base_dir + '/*.ini') or {
		return error('failed to list versions: ${err}')
	}
	log.debug('versions: ${versions}')

	versions.sort_with_compare(fn (a &string, b &string) int {
		a_parts := a.split_any('.-/')
		b_parts := b.split_any('.-/')

		for i in 0 .. math.min(a_parts.len, b_parts.len) {
			if a_parts[i] == b_parts[i] {
				continue
			}

			return compare_ints(a_parts[i].int(), b_parts[i].int())
		}

		return 0
	})

	log.debug('sorted versions: ${versions}')

	version_file := match version {
		'latest' {
			versions.last()
		}
		else {
			cand := os.join_path(base_dir, '${version}.ini')
			if cand in versions {
				cand
			} else {
				return error('version not found: ${version}')
			}
		}
	}

	return UChromiumVersion.from_file(version_file)
}

fn (u UChromium) download(version UChromiumVersion) !string {
	dest := u.cache.file('dl', os.base(version.url))!
	log.info('downloading ${version.url} to ${dest}')

	if os.is_file(dest) {
		log.info('file already exists')
		return dest
	}

	http.download_file(version.url, dest) or {
		return error('failed to download ${version.url}: ${err}')
	}

	return dest
}

fn (u UChromium) install(version string, install_to string) ! {
	u.set_repo()!
	v := u.find_version(version)!
	log.info('version ${version} found')
	dest := u.download(v)!
	os.chmod(dest, 0o777)!
	os.rm(install_to) or {}
	os.cp(dest, install_to)!

	return
}

fn compare_ints(a int, b int) int {
	return if a < b {
		-1
	} else if a > b {
		1
	} else {
		0
	}
}
